import enum
import SimpleITK as sitk


@enum.unique
class Interpolation(enum.Enum):
    """Interpolation techniques available in ITK.

    Example:
        >>> import torchio as tio
        >>> transform = tio.RandomAffine(image_interpolation='nearest')
    """
    #: Interpolates image intensity at a non-integer pixel position by copying the intensity for the nearest neighbor.
    NEAREST: str = 'sitkNearestNeighbor'

    #: Linearly interpolates image intensity at a non-integer pixel position.
    LINEAR: str = 'sitkLinear'

    #: Computes the B-spline interpolation weights over the support region of the B-spline.
    BSPLINE: str = 'sitkBSpline'

    GAUSSIAN: str = 'sitkGaussian'
    LABEL_GAUSSIAN: str = 'sitkLabelGaussian'

    HAMMING: str = 'sitkHammingWindowedSinc'
    COSINE: str = 'sitkCosineWindowedSinc'
    WELCH: str = 'sitkWelchWindowedSinc'
    LANCZOS: str = 'sitkLanczosWindowedSinc'
    BLACKMAN: str = 'sitkBlackmanWindowedSinc'


def get_sitk_interpolator(interpolation: str) -> int:
    if not isinstance(interpolation, str):
        message = (
            f'Interpolation must be a string, not {type(interpolation)}'
        )
        raise ValueError(message)
    string = getattr(Interpolation, interpolation.upper()).value
    return getattr(sitk, string)



from collections import defaultdict
from typing import Tuple, Optional, Union, Sequence, Dict

import torch
import numpy as np

from ....torchio import DATA
from ....data.subject import Subject
from ... import IntensityTransform, FourierTransform
from .. import RandomTransform


class RandomGhosting(RandomTransform, IntensityTransform):
    r"""Add random MRI ghosting artifact.

    Discrete "ghost" artifacts may occur along the phase-encode direction
    whenever the position or signal intensity of imaged structures within the
    field-of-view vary or move in a regular (periodic) fashion. Pulsatile flow
    of blood or CSF, cardiac motion, and respiratory motion are the most
    important patient-related causes of ghost artifacts in clinical MR imaging
    (from `mriquestions.com <http://mriquestions.com/why-discrete-ghosts.html>`_).

    Args:
        num_ghosts: Number of 'ghosts' :math:`n` in the image.
            If :attr:`num_ghosts` is a tuple :math:`(a, b)`, then
            :math:`n \sim \mathcal{U}(a, b) \cap \mathbb{N}`.
            If only one value :math:`d` is provided,
            :math:`n \sim \mathcal{U}(0, d) \cap \mathbb{N}`.
        axes: Axis along which the ghosts will be created. If
            :attr:`axes` is a tuple, the axis will be randomly chosen
            from the passed values. Anatomical labels may also be used (see
            :class:`~torchio.transforms.augmentation.RandomFlip`).
        intensity: Positive number representing the artifact strength
            :math:`s` with respect to the maximum of the :math:`k`-space.
            If ``0``, the ghosts will not be visible. If a tuple
            :math:`(a, b)` is provided then :math:`s \sim \mathcal{U}(a, b)`.
            If only one value :math:`d` is provided,
            :math:`s \sim \mathcal{U}(0, d)`.
        restore: Number between ``0`` and ``1`` indicating how much of the
            :math:`k`-space center should be restored after removing the planes
            that generate the artifact.
        p: Probability that this transform will be applied.
        keys: See :class:`~torchio.transforms.Transform`.

    .. note:: The execution time of this transform does not depend on the
        number of ghosts.
    """
    def __init__(
            self,
            num_ghosts: Union[int, Tuple[int, int]] = (4, 10),
            axes: Union[int, Tuple[int, ...]] = (0, 1, 2),
            intensity: Union[float, Tuple[float, float]] = (0.5, 1),
            restore: float = 0.02,
            p: float = 1,
            keys: Optional[Sequence[str]] = None,
            ):
        super().__init__(p=p, keys=keys)
        if not isinstance(axes, tuple):
            try:
                axes = tuple(axes)
            except TypeError:
                axes = (axes,)
        for axis in axes:
            if not isinstance(axis, str) and axis not in (0, 1, 2):
                raise ValueError(f'Axes must be in (0, 1, 2), not "{axes}"')
        self.axes = axes
        self.num_ghosts_range = self.parse_range(
            num_ghosts, 'num_ghosts', min_constraint=0, type_constraint=int)
        self.intensity_range = self.parse_range(
            intensity, 'intensity_range', min_constraint=0)
        self.restore = _parse_restore(restore)

    def apply_transform(self, subject: Subject) -> Subject:
        arguments = defaultdict(dict)
        if any(isinstance(n, str) for n in self.axes):
            subject.check_consistent_orientation()
        for name, image in self.get_images_dict(subject).items():
            is_2d = image.is_2d()
            axes = [a for a in self.axes if a != 2] if is_2d else self.axes
            params = self.get_params(
                self.num_ghosts_range,
                axes,
                self.intensity_range,
            )
            num_ghosts_param, axis_param, intensity_param = params
            arguments['num_ghosts'][name] = num_ghosts_param
            arguments['axis'][name] = axis_param
            arguments['intensity'][name] = intensity_param
            arguments['restore'][name] = self.restore
        transform = Ghosting(**arguments)
        transformed = transform(subject)
        return transformed

    def get_params(
            self,
            num_ghosts_range: Tuple[int, int],
            axes: Tuple[int, ...],
            intensity_range: Tuple[float, float],
            ) -> Tuple:
        ng_min, ng_max = num_ghosts_range
        num_ghosts = torch.randint(ng_min, ng_max + 1, (1,)).item()
        axis = axes[torch.randint(0, len(axes), (1,))]
        intensity = self.sample_uniform(*intensity_range).item()
        return num_ghosts, axis, intensity


class Ghosting(IntensityTransform, FourierTransform):
    r"""Add MRI ghosting artifact.

    Discrete "ghost" artifacts may occur along the phase-encode direction
    whenever the position or signal intensity of imaged structures within the
    field-of-view vary or move in a regular (periodic) fashion. Pulsatile flow
    of blood or CSF, cardiac motion, and respiratory motion are the most
    important patient-related causes of ghost artifacts in clinical MR imaging
    (from `mriquestions.com <http://mriquestions.com/why-discrete-ghosts.html>`_).

    Args:
        num_ghosts: Number of 'ghosts' :math:`n` in the image.
        axes: Axis along which the ghosts will be created.
        intensity: Positive number representing the artifact strength
            :math:`s` with respect to the maximum of the :math:`k`-space.
            If ``0``, the ghosts will not be visible.
        restore: Number between ``0`` and ``1`` indicating how much of the
            :math:`k`-space center should be restored after removing the planes
            that generate the artifact.
        keys: See :class:`~torchio.transforms.Transform`.

    .. note:: The execution time of this transform does not depend on the
        number of ghosts.
    """
    def __init__(
            self,
            num_ghosts: Union[int, Dict[str, int]],
            axis: Union[int, Dict[str, int]],
            intensity: Union[float, Dict[str, float]],
            restore: Union[float, Dict[str, float]],
            keys: Optional[Sequence[str]] = None,
            ):
        super().__init__(keys=keys)
        self.axis = axis
        self.num_ghosts = num_ghosts
        self.intensity = intensity
        self.restore = restore
        self.args_names = 'num_ghosts', 'axis', 'intensity', 'restore'

    def apply_transform(self, subject: Subject) -> Subject:
        axis = self.axis
        num_ghosts = self.num_ghosts
        intensity = self.intensity
        restore = self.restore
        for name, image in self.get_images_dict(subject).items():
            if self.arguments_are_dict():
                axis = self.axis[name]
                num_ghosts = self.num_ghosts[name]
                intensity = self.intensity[name]
                restore = self.restore[name]
            transformed_tensors = []
            for tensor in image.data:
                transformed_tensor = self.add_artifact(
                    tensor,
                    num_ghosts,
                    axis,
                    intensity,
                    restore,
                )
                transformed_tensors.append(transformed_tensor)
            image[DATA] = torch.stack(transformed_tensors)
        return subject

    def add_artifact(
            self,
            tensor: torch.Tensor,
            num_ghosts: int,
            axis: int,
            intensity: float,
            restore_center: float,
            ):
        if not num_ghosts or not intensity:
            return tensor

        array = tensor.numpy()
        spectrum = self.fourier_transform(array)

        shape = np.array(array.shape)
        ri, rj, rk = np.round(restore_center * shape).astype(np.uint16)
        mi, mj, mk = np.array(array.shape) // 2

        # Variable "planes" is the part of the spectrum that will be modified
        if axis == 0:
            planes = spectrum[::num_ghosts, :, :]
            restore = spectrum[mi, :, :].copy()
        elif axis == 1:
            planes = spectrum[:, ::num_ghosts, :]
            restore = spectrum[:, mj, :].copy()
        elif axis == 2:
            planes = spectrum[:, :, ::num_ghosts]
            restore = spectrum[:, :, mk].copy()

        # Multiply by 0 if intensity is 1
        planes *= 1 - intensity

        # Restore the center of k-space to avoid extreme artifacts
        if axis == 0:
            spectrum[mi, :, :] = restore
        elif axis == 1:
            spectrum[:, mj, :] = restore
        elif axis == 2:
            spectrum[:, :, mk] = restore

        array_ghosts = self.inv_fourier_transform(spectrum)
        array_ghosts = np.real(array_ghosts).astype(np.float32)
        return torch.from_numpy(array_ghosts)


def _parse_restore(restore):
    if not isinstance(restore, float):
        raise TypeError(f'Restore must be a float, not {restore}')
    if not 0 <= restore <= 1:
        message = (
            f'Restore must be a number between 0 and 1, not {restore}')
        raise ValueError(message)
    return restore



from torchio.transforms import RandomAffine
from ...utils import TorchioTestCase


class TestRandomAffine(TorchioTestCase):
    """Tests for `RandomAffine`."""
    def setUp(self):
        # Set image origin far from center
        super().setUp()
        affine = self.sample_subject.t1.affine
        affine[:3, 3] = 1e5

    def test_rotation_image(self):
        # Rotation around image center
        transform = RandomAffine(
            degrees=(90, 90),
            default_pad_value=0,
            center='image',
        )
        transformed = transform(self.sample_subject)
        total = transformed.t1.data.sum()
        self.assertNotEqual(total, 0)

    def test_rotation_origin(self):
        # Rotation around far away point, image should be empty
        transform = RandomAffine(
            degrees=(90, 90),
            default_pad_value=0,
            center='origin',
        )
        transformed = transform(self.sample_subject)
        total = transformed.t1.data.sum()
        self.assertEqual(total, 0)

    def test_no_rotation(self):
        transform = RandomAffine(
            scales=(1, 1),
            degrees=(0, 0),
            default_pad_value=0,
            center='image',
        )
        transformed = transform(self.sample_subject)
        self.assertTensorAlmostEqual(self.sample_subject.t1.data, transformed.t1.data)

        transform = RandomAffine(
            scales=(1, 1),
            degrees=(180, 180),
            default_pad_value=0,
            center='image',
        )
        transformed = transform(self.sample_subject)
        transformed = transform(transformed)
        self.assertTensorAlmostEqual(self.sample_subject.t1.data, transformed.t1.data)

    def test_translation(self):
        transform = RandomAffine(
            scales=(1, 1),
            degrees=0,
            translation=(5, 5)
        )
        transformed = transform(self.sample_subject)

        # I think the right test should be the following one:
        # self.assertTensorAlmostEqual(
        #     self.sample_subject.t1.data[:, :-5, :-5, :-5],
        #     transformed.t1.data[:, 5:, 5:, 5:]
        # )

        # However the passing test is this one:
        self.assertTensorAlmostEqual(
            self.sample_subject.t1.data[:, :-5, :-5, 5:],
            transformed.t1.data[:, 5:, 5:, :-5]
        )

    def test_negative_scales(self):
        with self.assertRaises(ValueError):
            RandomAffine(scales=(-1, 1))

    def test_scale_too_large(self):
        with self.assertRaises(ValueError):
            RandomAffine(scales=1.5)

    def test_scales_range_with_negative_min(self):
        with self.assertRaises(ValueError):
            RandomAffine(scales=(-1, 4))

    def test_wrong_scales_type(self):
        with self.assertRaises(ValueError):
            RandomAffine(scales='wrong')

    def test_wrong_degrees_type(self):
        with self.assertRaises(ValueError):
            RandomAffine(degrees='wrong')

    def test_too_many_translation_values(self):
        with self.assertRaises(ValueError):
            RandomAffine(translation=(-10, 4, 42))

    def test_wrong_translation_type(self):
        with self.assertRaises(ValueError):
            RandomAffine(translation='wrong')

    def test_wrong_center(self):
        with self.assertRaises(ValueError):
            RandomAffine(center=0)

    def test_wrong_default_pad_value(self):
        with self.assertRaises(ValueError):
            RandomAffine(default_pad_value='wrong')

    def test_wrong_image_interpolation_type(self):
        with self.assertRaises(TypeError):
            RandomAffine(image_interpolation=0)

    def test_wrong_image_interpolation_value(self):
        with self.assertRaises(ValueError):
            RandomAffine(image_interpolation='wrong')

    def test_incompatible_args_isotropic(self):
        with self.assertRaises(ValueError):
            RandomAffine(scales=(0.8, 0.5, 0.1), isotropic=True)

    def test_parse_scales(self):
        def do_assert(transform):
            self.assertEqual(transform.scales, 3 * (0.9, 1.1))
        do_assert(RandomAffine(scales=0.1))
        do_assert(RandomAffine(scales=(0.9, 1.1)))
        do_assert(RandomAffine(scales=3 * (0.1,)))
        do_assert(RandomAffine(scales=3 * [0.9, 1.1]))

    def test_parse_degrees(self):
        def do_assert(transform):
            self.assertEqual(transform.degrees, 3 * (-10, 10))
        do_assert(RandomAffine(degrees=10))
        do_assert(RandomAffine(degrees=(-10, 10)))
        do_assert(RandomAffine(degrees=3 * (10,)))
        do_assert(RandomAffine(degrees=3 * [-10, 10]))

    def test_parse_translation(self):
        def do_assert(transform):
            self.assertEqual(transform.translation, 3 * (-10, 10))
        do_assert(RandomAffine(translation=10))
        do_assert(RandomAffine(translation=(-10, 10)))
        do_assert(RandomAffine(translation=3 * (10,)))
        do_assert(RandomAffine(translation=3 * [-10, 10]))



import torch
import torchio as tio
from ...utils import TorchioTestCase


class TestZNormalization(TorchioTestCase):
    """Tests for :class:`ZNormalization` class."""

    def test_z_normalization(self):
        transform = tio.ZNormalization()
        transformed = transform(self.sample_subject)
        self.assertAlmostEqual(float(transformed.t1.data.mean()), 0., places=6)
        self.assertAlmostEqual(float(transformed.t1.data.std()), 1.)

    def test_no_std(self):
        image = tio.ScalarImage(tensor=torch.ones(1, 2, 2, 2))
        with self.assertRaises(RuntimeError):
            tio.ZNormalization()(image)



# Use duecredit to provide a citation to relevant work to
# be cited. This does nothing unless the user has duecredit installed
# and calls this with duecredit (as in `python -m duecredit script.py`):
from .external.due import due, Doi, BibTeX

BIBTEX = r"""@misc{fern2020torchio,
   title={TorchIO: a Python library for efficient loading, preprocessing, augmentation and patch-based sampling of medical images in deep learning},
   author={Fernando Pérez-García and Rachel Sparks and Sebastien Ourselin},
   year={2020},
   eprint={2003.04696},
   archivePrefix={arXiv},
   primaryClass={eess.IV}
} """

TITLE = (
    'TorchIO: a Python library for efficient loading, preprocessing,'
    ' augmentation and patch-based sampling of medical images in deep learning'
)

DESCRIPTION = (
    'Tools for loading, augmenting and writing 3D medical images'
    ' on PyTorch'
)

due.cite(
    BibTeX(BIBTEX),
    description=TITLE,
    path='torchio',
    cite_module=True,
)

due.cite(
    Doi('10.5281/zenodo.3739230'),
    description=DESCRIPTION,
    path='torchio',
    tags=['implementation'],
    cite_module=True,
)



from collections import defaultdict
from typing import Tuple, Optional, Union, Dict, Sequence

import torch
from ....torchio import DATA
from ....data.subject import Subject
from ... import IntensityTransform
from .. import RandomTransform


class RandomNoise(RandomTransform, IntensityTransform):
    r"""Add Gaussian noise with random parameters.

    Add noise sampled from a normal distribution with random parameters.

    Args:
        mean: Mean :math:`\mu` of the Gaussian distribution
            from which the noise is sampled.
            If two values :math:`(a, b)` are provided,
            then :math:`\mu \sim \mathcal{U}(a, b)`.
            If only one value :math:`d` is provided,
            :math:`\mu \sim \mathcal{U}(-d, d)`.
        std: Standard deviation :math:`\sigma` of the Gaussian distribution
            from which the noise is sampled.
            If two values :math:`(a, b)` are provided,
            then :math:`\sigma \sim \mathcal{U}(a, b)`.
            If only one value :math:`d` is provided,
            :math:`\sigma \sim \mathcal{U}(0, d)`.
        p: Probability that this transform will be applied.
        keys: See :class:`~torchio.transforms.Transform`.
    """
    def __init__(
            self,
            mean: Union[float, Tuple[float, float]] = 0,
            std: Union[float, Tuple[float, float]] = (0, 0.25),
            p: float = 1,
            keys: Optional[Sequence[str]] = None,
            ):
        super().__init__(p=p, keys=keys)
        self.mean_range = self.parse_range(mean, 'mean')
        self.std_range = self.parse_range(std, 'std', min_constraint=0)

    def apply_transform(self, subject: Subject) -> Subject:
        arguments = defaultdict(dict)
        for image_name in self.get_images_dict(subject):
            mean, std, seed = self.get_params(self.mean_range, self.std_range)
            arguments['mean'][image_name] = mean
            arguments['std'][image_name] = std
            arguments['seed'][image_name] = seed
        transform = Noise(**arguments)
        transformed = transform(subject)
        return transformed

    def get_params(
            self,
            mean_range: Tuple[float, float],
            std_range: Tuple[float, float],
            ) -> Tuple[float, float]:
        mean = self.sample_uniform(*mean_range).item()
        std = self.sample_uniform(*std_range).item()
        seed = self.get_random_seed()
        return mean, std, seed


class Noise(IntensityTransform):
    r"""Add Gaussian noise.

    Add noise sampled from a normal distribution.

    Args:
        mean: Mean :math:`\mu` of the Gaussian distribution
            from which the noise is sampled.
        std: Standard deviation :math:`\sigma` of the Gaussian distribution
            from which the noise is sampled.
        seed: Seed for the random number generator.
        keys: See :class:`~torchio.transforms.Transform`.
    """
    def __init__(
            self,
            mean: Union[float, Dict[str, float]],
            std: Union[float, Dict[str, float]],
            seed: Union[int, Sequence[int]],
            keys: Optional[Sequence[str]] = None,
            ):
        super().__init__(keys=keys)
        self.mean = mean
        self.std = std
        self.seed = seed
        self.invert_transform = False
        self.args_names = 'mean', 'std', 'seed'

    def apply_transform(self, subject: Subject) -> Subject:
        args = self.mean, self.std, self.seed
        for name, image in self.get_images_dict(subject).items():
            if self.arguments_are_dict():
                mean, std, seed = [arg[name] for arg in args]
            with self._use_seed(seed):
                noise = get_noise(image[DATA], mean, std)
            if self.invert_transform:
                noise *= -1
            image[DATA] = image[DATA] + noise
        return subject


def get_noise(tensor: torch.Tensor, mean: float, std: float) -> torch.Tensor:
    return torch.randn(*tensor.shape) * std + mean



import numpy as np
from torchio.transforms import ToCanonical
from ...utils import TorchioTestCase


class TestToCanonical(TorchioTestCase):
    """Tests for :class:`ToCanonical` class."""

    def test_no_changes(self):
        transform = ToCanonical()
        transformed = transform(self.sample_subject)
        self.assertTensorEqual(transformed.t1.data, self.sample_subject.t1.data)
        self.assertTensorEqual(transformed.t1.affine, self.sample_subject.t1.affine)

    def test_las_to_ras(self):
        self.sample_subject.t1.affine[0, 0] = -1    # Change orientation to 'LAS'
        transform = ToCanonical()
        transformed = transform(self.sample_subject)
        self.assertEqual(transformed.t1.orientation, ('R', 'A', 'S'))
        self.assertTensorEqual(
            transformed.t1.data,
            self.sample_subject.t1.data.numpy()[:, ::-1, :, :]
        )

        fixture = np.eye(4)
        fixture[0, -1] = -self.sample_subject.t1.spatial_shape[0] + 1
        self.assertTensorEqual(transformed.t1.affine, fixture)



from numbers import Number
from typing import Union, Sequence, Optional

import numpy as np
import nibabel as nib
import torch

from ....torchio import DATA, AFFINE
from ....data.subject import Subject
from .bounds_transform import BoundsTransform, TypeBounds


class Pad(BoundsTransform):
    r"""Pad an image.

    Args:
        padding: Tuple
            :math:`(w_{ini}, w_{fin}, h_{ini}, h_{fin}, d_{ini}, d_{fin})`
            defining the number of values padded to the edges of each axis.
            If the initial shape of the image is
            :math:`W \times H \times D`, the final shape will be
            :math:`(w_{ini} + W + w_{fin}) \times (h_{ini} + H + h_{fin})
            \times (d_{ini} + D + d_{fin})`.
            If only three values :math:`(w, h, d)` are provided, then
            :math:`w_{ini} = w_{fin} = w`,
            :math:`h_{ini} = h_{fin} = h` and
            :math:`d_{ini} = d_{fin} = d`.
            If only one value :math:`n` is provided, then
            :math:`w_{ini} = w_{fin} = h_{ini} = h_{fin} =
            d_{ini} = d_{fin} = n`.
        padding_mode: See possible modes in `NumPy docs`_. If it is a number,
            the mode will be set to ``'constant'``.
        p: Probability that this transform will be applied.
        keys: See :class:`~torchio.transforms.Transform`.

    .. _NumPy docs: https://numpy.org/doc/stable/reference/generated/numpy.pad.html
    """

    PADDING_MODES = (
        'empty',
        'edge',
        'wrap',
        'constant',
        'linear_ramp',
        'maximum',
        'mean',
        'median',
        'minimum',
        'reflect',
        'symmetric',
    )

    def __init__(
            self,
            padding: TypeBounds,
            padding_mode: Union[str, float] = 0,
            p: float = 1,
            keys: Optional[Sequence[str]] = None,
            ):
        super().__init__(padding, p=p, keys=keys)
        self.padding = padding
        self.padding_mode, self.fill = self.parse_padding_mode(padding_mode)
        self.args_names = 'padding', 'padding_mode'

    @classmethod
    def parse_padding_mode(cls, padding_mode):
        if padding_mode in cls.PADDING_MODES:
            fill = None
        elif isinstance(padding_mode, Number):
            fill = padding_mode
            padding_mode = 'constant'
        else:
            message = (
                f'Padding mode "{padding_mode}" not valid. Valid options are'
                f' {list(cls.PADDING_MODES)} or a number'
            )
            raise KeyError(message)
        return padding_mode, fill

    def apply_transform(self, subject: Subject) -> Subject:
        low = self.bounds_parameters[::2]
        for image in self.get_images(subject):
            new_origin = nib.affines.apply_affine(image.affine, -np.array(low))
            new_affine = image.affine.copy()
            new_affine[:3, 3] = new_origin
            kwargs = {'mode': self.padding_mode}
            if self.padding_mode == 'constant':
                kwargs['constant_values'] = self.fill
            pad_params = self.bounds_parameters
            paddings = (0, 0), pad_params[:2], pad_params[2:4], pad_params[4:]
            padded = np.pad(image[DATA], paddings, **kwargs)
            image[DATA] = torch.from_numpy(padded)
            image[AFFINE] = new_affine
        return subject

    def inverse(self):
        from .crop import Crop
        return Crop(self.padding)



import torch
from torchio.transforms import RandomElasticDeformation
from ...utils import TorchioTestCase


class TestRandomElasticDeformation(TorchioTestCase):
    """Tests for `RandomElasticDeformation`."""

    def test_random_elastic_deformation(self):
        transform = RandomElasticDeformation(
            num_control_points=5,
            max_displacement=(2, 3, 5),  # half grid spacing is (3.3, 3.3, 5)
        )
        keys = ('t1', 't2', 'label')
        fixtures = 2916.7192, 2955.1265, 2950
        torch_rng_state = torch.random.get_rng_state()
        torch.manual_seed(42)
        transformed = transform(self.sample_subject)
        torch.random.set_rng_state(torch_rng_state)
        for key, fixture in zip(keys, fixtures):
            sample_data = self.sample_subject[key].numpy()
            transformed_data = transformed[key].numpy()
            transformed_total = transformed_data.sum()
            # Make sure that intensities have changed
            self.assertTensorNotEqual(sample_data, transformed_data)
            self.assertAlmostEqual(transformed_total, fixture, places=4)

    def test_inputs_pta_gt_one(self):
        with self.assertRaises(ValueError):
            RandomElasticDeformation(p=1.5)

    def test_inputs_pta_lt_zero(self):
        with self.assertRaises(ValueError):
            RandomElasticDeformation(p=-1)

    def test_inputs_interpolation_int(self):
        with self.assertRaises(TypeError):
            RandomElasticDeformation(image_interpolation=1)

    def test_inputs_interpolation(self):
        with self.assertRaises(TypeError):
            RandomElasticDeformation(image_interpolation=0)

    def test_num_control_points_noint(self):
        with self.assertRaises(ValueError):
            RandomElasticDeformation(num_control_points=2.5)

    def test_num_control_points_small(self):
        with self.assertRaises(ValueError):
            RandomElasticDeformation(num_control_points=3)

    def test_max_displacement_no_num(self):
        with self.assertRaises(ValueError):
            RandomElasticDeformation(max_displacement=None)

    def test_max_displacement_negative(self):
        with self.assertRaises(ValueError):
            RandomElasticDeformation(max_displacement=-1)

    def test_wrong_locked_borders(self):
        with self.assertRaises(ValueError):
            RandomElasticDeformation(locked_borders=-1)

    def test_coarse_grid_removed(self):
        with self.assertRaises(ValueError):
            RandomElasticDeformation(
                num_control_points=(4, 5, 6),
                locked_borders=2,
            )

    def test_folding(self):
        # Assume shape is (10, 20, 30) and spacing is (1, 1, 1)
        # Then grid spacing is (10/(12-2), 20/(5-2), 30/(5-2))
        # or (1, 6.7, 10), and half is (0.5, 3.3, 5)
        transform = RandomElasticDeformation(
            num_control_points=(12, 5, 5),
            max_displacement=6,
        )
        with self.assertWarns(RuntimeWarning):
            transform(self.sample_subject)

    def test_num_control_points(self):
        RandomElasticDeformation(num_control_points=5)
        RandomElasticDeformation(num_control_points=(5, 6, 7))

    def test_max_displacement(self):
        RandomElasticDeformation(max_displacement=5)
        RandomElasticDeformation(max_displacement=(5, 6, 7))

    def test_no_displacement(self):
        transform = RandomElasticDeformation(max_displacement=0)
        transformed = transform(self.sample_subject)
        self.assertTensorEqual(
            self.sample_subject.t1.data,
            transformed.t1.data,
        )
        self.assertTensorEqual(
            self.sample_subject.label.data,
            transformed.label.data,
        )



import urllib.parse
from ...utils import get_torchio_cache_dir, download_and_extract_archive
from ... import ScalarImage, LabelMap, DATA
from .mni import SubjectMNI


TISSUES_2008 = {
    1: 'Cerebro-spinal fluid',
    2: 'Gray Matter',
    3: 'White Matter',
    4: 'Fat',
    5: 'Muscles',
    6: 'Skin and Muscles',
    7: 'Skull',
    9: 'Fat 2',
    10: 'Dura',
    11: 'Marrow',
    12: 'Vessels',
}


class Colin27(SubjectMNI):
    r"""Colin27 MNI template.

    More information can be found in the website of the
    `1998 <http://nist.mni.mcgill.ca/?p=935>`_ and
    `2008 <http://www.bic.mni.mcgill.ca/ServicesAtlases/Colin27Highres>`_
    versions.

    .. image:: http://www.bic.mni.mcgill.ca/uploads/ServicesAtlases/mni_colin27_2008.jpg
        :alt: MNI Colin 27 2008 version

    Arguments:
        version: Template year. It can be ``1998`` or ``2008``.

    .. warning:: The resolution of the ``2008`` version is quite high. The
        subject instance will contain four images of size
        :math:`362 \times 434 \times 362`, therefore applying a transform to
        it might take longer than expected.

    Example:
        >>> import torchio as tio
        >>> colin_1998 = tio.datasets.Colin27(version=1998)
        >>> colin_1998
        Colin27(Keys: ('t1', 'head', 'brain'); images: 3)
        >>> colin_1998.load()
        >>> colin_1998.t1
        ScalarImage(shape: (1, 181, 217, 181); spacing: (1.00, 1.00, 1.00); orientation: RAS+; memory: 27.1 MiB; type: intensity)
        >>>
        >>> colin_2008 = tio.datasets.Colin27(version=2008)
        >>> colin_2008
        Colin27(Keys: ('t1', 't2', 'pd', 'cls'); images: 4)
        >>> colin_2008.load()
        >>> colin_2008.t1
        ScalarImage(shape: (1, 362, 434, 362); spacing: (0.50, 0.50, 0.50); orientation: RAS+; memory: 217.0 MiB; type: intensity)

    """
    def __init__(self, version=1998):
        if version not in (1998, 2008):
            raise ValueError(f'Version must be 1998 or 2008, not "{version}"')
        self.name = f'mni_colin27_{version}_nifti'
        self.url_dir = urllib.parse.urljoin(self.url_base, 'colin27/')
        self.filename = f'{self.name}.zip'
        self.url = urllib.parse.urljoin(self.url_dir, self.filename)
        download_root = get_torchio_cache_dir() / self.name
        if not download_root.is_dir():
            download_and_extract_archive(
                self.url,
                download_root=download_root,
                filename=self.filename,
            )
            # Fix label map (https://github.com/fepegar/torchio/issues/220)
            if version == 2008:
                path = download_root / 'colin27_cls_tal_hires.nii'
                cls_image = LabelMap(path)
                cls_image[DATA] = cls_image[DATA].round().byte()
                cls_image.save(path)

        if version == 1998:
            t1, head, mask = [
                download_root / f'colin27_t1_tal_lin{suffix}.nii'
                for suffix in ('', '_headmask', '_mask')
            ]
            super().__init__(
                t1=ScalarImage(t1),
                head=LabelMap(head),
                brain=LabelMap(mask),
            )
        elif version == 2008:
            t1, t2, pd, label = [
                download_root / f'colin27_{name}_tal_hires.nii'
                for name in ('t1', 't2', 'pd', 'cls')
            ]
            super().__init__(
                t1=ScalarImage(t1),
                t2=ScalarImage(t2),
                pd=ScalarImage(pd),
                cls=LabelMap(label, labels=TISSUES_2008),
            )



from .transform import Transform
from .fourier import FourierTransform
from .spatial_transform import SpatialTransform
from .intensity_transform import IntensityTransform

# Generic
from .lambda_transform import Lambda

# Augmentation
from .augmentation.composition import OneOf
from .augmentation.composition import Compose

from .augmentation.spatial import RandomFlip, Flip
from .augmentation.spatial import RandomAffine, Affine
from .augmentation.spatial import RandomAnisotropy
from .augmentation.spatial import RandomElasticDeformation, ElasticDeformation

from .augmentation.intensity import RandomSwap, Swap
from .augmentation.intensity import RandomBlur, Blur
from .augmentation.intensity import RandomNoise, Noise
from .augmentation.intensity import RandomSpike, Spike
from .augmentation.intensity import RandomGamma, Gamma
from .augmentation.intensity import RandomMotion, Motion
from .augmentation.intensity import RandomGhosting, Ghosting
from .augmentation.intensity import RandomBiasField, BiasField
from .augmentation.intensity import RandomLabelsToImage, LabelsToImage

# Preprocessing
from .preprocessing import Pad
from .preprocessing import Crop
from .preprocessing import Resample
from .preprocessing import CropOrPad
from .preprocessing import ToCanonical
from .preprocessing import ZNormalization
from .preprocessing import RescaleIntensity
from .preprocessing import HistogramStandardization
from .preprocessing.intensity.histogram_standardization import train as train_histogram


__all__ = [
    'Transform',
    'FourierTransform',
    'SpatialTransform',
    'IntensityTransform',
    'Lambda',
    'OneOf',
    'Compose',
    'RandomFlip',
    'Flip',
    'RandomAffine',
    'Affine',
    'RandomAnisotropy',
    'RandomElasticDeformation',
    'ElasticDeformation',
    'RandomSwap',
    'Swap',
    'RandomBlur',
    'Blur',
    'RandomNoise',
    'Noise',
    'RandomSpike',
    'Spike',
    'RandomGamma',
    'Gamma',
    'RandomMotion',
    'Motion',
    'RandomGhosting',
    'Ghosting',
    'RandomBiasField',
    'BiasField',
    'RandomLabelsToImage',
    'LabelsToImage',
    'Pad',
    'Crop',
    'Resample',
    'ToCanonical',
    'ZNormalization',
    'HistogramStandardization',
    'RescaleIntensity',
    'CropOrPad',
    'train_histogram',
]



from .spatial.pad import Pad
from .spatial.crop import Crop
from .spatial.resample import Resample
from .spatial.crop_or_pad import CropOrPad
from .spatial.to_canonical import ToCanonical

from .intensity.rescale import RescaleIntensity
from .intensity.z_normalization import ZNormalization
from .intensity.histogram_standardization import HistogramStandardization


__all__ = [
    'Pad',
    'Crop',
    'Resample',
    'ToCanonical',
    'CropOrPad',
    'RescaleIntensity',
    'ZNormalization',
    'HistogramStandardization',
]
